/* $Id: semaphor.h,v 1.2 2002/04/26 23:09:21 smilcke Exp $ */

#ifndef _I386_SEMAPHORE_H
#define _I386_SEMAPHORE_H

/*
 * SMP- and interrupt-safe semaphores..
 *
 * (C) Copyright 1996 Linus Torvalds
 *
 * Modified 1996-12-23 by Dave Grothe <dave@gcom.com> to fix bugs in
 *                     the original code and to make semaphore waits
 *                     interruptible so that processes waiting on
 *                     semaphores can be killed.
 * Modified 1999-02-14 by Andrea Arcangeli, split the sched.c helper
 *		       functions in asm/sempahore-helper.h while fixing a
 *		       potential and subtle race discovered by Ulrich Schmid
 *		       in down_interruptible(). Since I started to play here I
 *		       also implemented the `trylock' semaphore operation.
 *          1999-07-02 Artur Skawina <skawina@geocities.com>
 *                     Optimized "0(ecx)" -> "(ecx)" (the assembler does not
 *                     do this). Changed calling sequences from push/jmp to
 *                     traditional call/ret.
 *
 * If you would like to see an analysis of this implementation, please
 * ftp to gcom.com and download the file
 * /pub/linux/src/semaphore/semaphore-2.0.24.tar.gz.
 *
 */

#include <asm/atomic.h>
#include <linux/spinlock.h>

struct semaphore {
	atomic_t count;
	int sleepers;
	void * wait;
#if WAITQUEUE_DEBUG
	long __magic;
#endif
};

#if WAITQUEUE_DEBUG
# define __SEM_DEBUG_INIT(name) \
		, (int)&(name).__magic
#else
# define __SEM_DEBUG_INIT(name)
#endif

#define __SEMAPHORE_INITIALIZER(name,count) \
{ ATOMIC_INIT(count), 0, __WAIT_QUEUE_HEAD_INITIALIZER((name).wait) \
	__SEM_DEBUG_INIT(name) }

#define __MUTEX_INITIALIZER(name) \
	__SEMAPHORE_INITIALIZER(name,1)

#define __DECLARE_SEMAPHORE_GENERIC(name,count) \
	struct semaphore name = __SEMAPHORE_INITIALIZER(name,count)

#define DECLARE_MUTEX(name) __DECLARE_SEMAPHORE_GENERIC(name,1)
#define DECLARE_MUTEX_LOCKED(name) __DECLARE_SEMAPHORE_GENERIC(name,0)

extern void sema_init (struct semaphore *sem, int val);
#if (defined(TARGET_OS2) && !defined(NOOS2LXAPI))
extern void (*init_MUTEX)(struct semaphore *sem);
extern void (*init_MUTEX_LOCKED)(struct semaphore *sem);
#else
void init_MUTEX (struct semaphore *sem);
void init_MUTEX_LOCKED (struct semaphore *sem);
#endif

 void __down_failed(void /* special register calling convention */);
 int  __down_failed_interruptible(void  /* params in registers */);
 int  __down_failed_trylock(void  /* params in registers */);
 void __up_wakeup(void /* special register calling convention */);

 void __down(struct semaphore * sem);
 int  __down_interruptible(struct semaphore * sem);
 int  __down_trylock(struct semaphore * sem);
 void __up(struct semaphore * sem);

/*
 * This is ugly, but we want the default case to fall through.
 * "down_failed" is a special asm handler that calls the C
 * routine that actually waits. See arch/i386/lib/semaphore.S
 */
#if (defined(TARGET_OS2) && !defined(NOOS2LXAPI))
extern void (*down)(struct semaphore *sem);
#else
extern void down(struct semaphore * sem);
#endif

extern int down_interruptible(struct semaphore * sem);

extern int down_trylock(struct semaphore * sem);

/*
 * Note! This is subtle. We jump to wake people up only if
 * the semaphore was negative (== somebody was waiting on it).
 * The default case (no contention) will result in NO
 * jumps for both down() and up().
 */
#if (defined(TARGET_OS2) && !defined(NOOS2LXAPI))
extern void (*up)(struct semaphore *sem);
#else
extern void up(struct semaphore * sem);
#endif

#endif
